# The LearningOnline Network with CAPA
# Search Catalog
#
# $Id: lonsearchcat.pm,v 1.135 2002/07/03 20:10:14 matthew Exp $
#
# Copyright Michigan State University Board of Trustees
#
# This file is part of the LearningOnline Network with CAPA (LON-CAPA).
#
# LON-CAPA is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# LON-CAPA is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with LON-CAPA; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# /home/httpd/html/adm/gpl.txt
#
# http://www.lon-capa.org/
#
# YEAR=2001
# 3/8, 3/12, 3/13, 3/14, 3/15, 3/19 Scott Harrison
# 3/20, 3/21, 3/22, 3/26, 3/27, 4/2, 8/15, 8/24, 8/25 Scott Harrison
# 10/12,10/14,10/15,10/16,11/28,11/29,12/10,12/12,12/16 Scott Harrison
# YEAR=2002
# 1/17 Scott Harrison
# 6/17 Matthew Hall
#
###############################################################################
###############################################################################

=pod 

=head1 NAME

lonsearchcat

=head1 SYNOPSIS

Search interface to LON-CAPAs digital library

=head1 DESCRIPTION

This module enables searching for a distributed browseable catalog.

This is part of the LearningOnline Network with CAPA project
described at http://www.lon-capa.org.

lonsearchcat presents the user with an interface to search the LON-CAPA
digital library.  lonsearchcat also initiates the execution of a search
by sending the search parameters to LON-CAPA servers.  The progress of 
search (on a server basis) is displayed to the user in a seperate window.

=head1 Internals

=over 4

=cut

###############################################################################
###############################################################################

###############################################################################
##                                                                           ##
## ORGANIZATION OF THIS PERL MODULE                                          ##
##                                                                           ##
## 1. Modules used by this module                                            ##
## 2. Variables used throughout the module                                   ##
## 3. handler subroutine called via Apache and mod_perl                      ##
## 4. Other subroutines                                                      ##
##                                                                           ##
###############################################################################

package Apache::lonsearchcat;

# ------------------------------------------------- modules used by this module
use strict;
use Apache::Constants qw(:common);
use Apache::lonnet();
use Apache::File();
use CGI qw(:standard);
use Text::Query;
use GDBM_File;
use Apache::loncommon();

# ---------------------------------------- variables used throughout the module

######################################################################
######################################################################

=pod 

=item Global variables

=over 4

=item $closebutton

button that closes the search window

=item $importbutton

button to take the select results and go to group sorting

=item %hash   

The ubiquitous database hash

=item $diropendb 

The full path to the (temporary) search database file.  This is set and
used in &handler() and is also used in &output_results().

=back 

=cut

######################################################################
######################################################################

# -- dynamically rendered interface components
my $closebutton;  # button that closes the search window
my $importbutton; # button to take the selected results and go to group sorting

# -- miscellaneous variables
my %hash;     # database hash
my $diropendb = "";    # db file

######################################################################
######################################################################

=pod 

=item &handler() - main handler invoked by httpd child

=item Variables

=over 4

=item $hidden

holds 'hidden' html forms

=item $scrout

string that holds portions of the screen output

=back 

=cut

######################################################################
######################################################################
sub handler {
    my $r = shift;
    untie %hash;

    $r->content_type('text/html');
    $r->send_http_header;
    return OK if $r->header_only;

    my $domain  = $r->dir_config('lonDefDomain');
    $diropendb= "/home/httpd/perl/tmp/".&Apache::lonnet::escape($domain).
            "\_".&Apache::lonnet::escape($ENV{'user.name'})."_searchcat.db";

    &Apache::loncommon::get_unprocessed_cgi($ENV{'QUERY_STRING'},
             ['catalogmode','launch','acts','mode','form','element',
              'reqinterface']);
    ##
    ## Clear out old values from database
    ##
    if ($ENV{'form.launch'} eq '1') {
	if (tie(%hash,'GDBM_File',$diropendb,&GDBM_WRCREAT,0640)) {
	    &start_fresh_session();
	    untie %hash;
	} else {
	    $r->print('<html><head></head><body>Unable to tie hash to db '.
		      'file</body></html>');
	    return OK;
	}
    }
    ##
    ## Produce some output, so people know it is working
    ##
    $r->print("\n");
    $r->rflush;
    ##
    ## Configure dynamic components of interface
    ##
    my $hidden;       # Holds 'hidden' html forms
    if ($ENV{'form.catalogmode'} eq 'interactive') {
	$hidden="<input type='hidden' name='catalogmode' value='interactive'>".
	    "\n";
        $closebutton="<input type='button' name='close' value='CLOSE' ".
	    "onClick='self.close()'>"."\n";
    } elsif ($ENV{'form.catalogmode'} eq 'groupsearch') {
	$hidden=<<END;
<input type='hidden' name='catalogmode' value='groupsearch'>
END
        $closebutton=<<END;
<input type='button' name='close' value='CLOSE' onClick='self.close()'>
END
        $importbutton=<<END;
<input type='button' name='import' value='IMPORT'
onClick='javascript:select_group()'>
END
    }
    $hidden .= &make_persistent({ "form.mode"    => $ENV{'form.mode'},
                                  "form.form"    => $ENV{'form.form'},
                                  "form.element" => $ENV{'form.element'},
                                  "form.date"    => 2 });
    ##
    ##  What are we doing?
    ##
    my $searchtype;
    $searchtype = 'Basic'    if ($ENV{'form.basicsubmit'}    eq 'SEARCH');
    $searchtype = 'Advanced' if ($ENV{'form.advancedsubmit'} eq 'SEARCH');
    if ($searchtype) {
        my ($query,$customquery,$customshow,$libraries) = 
            (undef,undef,undef,undef);
        if ($searchtype eq 'Basic') {
            $query = &parse_basic_search($r);
        } elsif ($ENV{'form.advancedsubmit'} eq 'SEARCH') {
            ($query,$customquery,$customshow,$libraries) 
                = &parse_advanced_search($r);
            return OK if (! defined($query));
        }
        # Send query statements over the network to be processed by 
        # either the SQL database or a recursive scheme of 'grep'-like 
        # actions (for custom metadata).
        $r->rflush();
        my $reply=&Apache::lonnet::metadata_query($query,$customquery,
                                               $customshow,$libraries);
        &output_results($searchtype,$r,$reply,$hidden);
    } elsif ($ENV{'form.reqinterface'} eq 'advanced') {
        # Output the advanced interface
        $r->print(&advanced_search_form($closebutton,$hidden));
    } else { 
        # Output normal search interface
        $r->print(&basic_search_form($closebutton,$hidden));
    }
    return OK;
} 

######################################################################
######################################################################

=pod 

=item &basic_search_form() 

Returns a scalar which holds html for the basic search form.

=cut

######################################################################
######################################################################

sub basic_search_form{
    my ($closebutton,$hidden) = @_;
    my $scrout=<<"ENDDOCUMENT";
<html>
<head>
<title>The LearningOnline Network with CAPA</title>
<script type="text/javascript">
    function openhelp(val) {
	openhelpwin=open('/adm/help/searchcat.html','helpscreen',
	     'scrollbars=1,width=600,height=300');
	openhelpwin.focus();
    }
</script>
</head>
<body bgcolor="#FFFFFF">
<img align='right' src='/adm/lonIcons/lonlogos.gif' />
<h1>Search Catalog</h1>
<form method="post" action="/adm/searchcat">
$hidden
<h3>Basic Search</h3>
<p>
Enter terms or phrases separated by AND, OR, or NOT 
then press SEARCH below.
</p>
<p>
<table>
<tr><td>
ENDDOCUMENT
    $scrout.='&nbsp;'.&simpletextfield('basicexp',$ENV{'form.basicexp'},40).
        '&nbsp;';
#    $scrout.=&simplecheckbox('allversions',$ENV{'form.allversions'});
#    $scrout.='<font color="#800000">Search historic archives</font>';
    $scrout.=<<ENDDOCUMENT;
</td><td><a href="/adm/searchcat?reqinterface=advanced">Advanced Search</a></td></tr></table>
</p>
<p>
&nbsp;<input type="submit" name="basicsubmit" value='SEARCH' />&nbsp;
$closebutton
<!-- view selection -->
<select name="viewselect">
<option value="Detailed Citation View" selected="true" >Detailed Citation View</option>
<option value="Summary View"  >Summary View</option>
<option value="Fielded Format">Fielded Format</option>
<option value="XML/SGML"      >XML/SGML</option>
</select>
<!-- end of view selection -->
<input type="button" value="HELP" onClick="openhelp()" />
</p>
</form>
</body>
</html>
ENDDOCUMENT
    return $scrout;
}
######################################################################
######################################################################

=pod 

=item &advanced_search_form() 

Returns a scalar which holds html for the advanced search form.

=cut

######################################################################
######################################################################

sub advanced_search_form{
    my ($closebutton,$hidden) = @_;
    my $advanced_buttons = <<"END";
<p>
<input type="submit" name="advancedsubmit" value='SEARCH' />
<input type="reset" name="reset" value='RESET' />
$closebutton
<input type="button" value="HELP" onClick="openhelp()" />
</p>
END
    my $scrout=<<"ENDHEADER";
<html>
<head>
<title>The LearningOnline Network with CAPA</title>
<script type="text/javascript">
    function openhelp(val) {
	openhelpwin=open('/adm/help/searchcat.html','helpscreen',
	     'scrollbars=1,width=600,height=300');
	openhelpwin.focus();
    }
</script>
</head>
<body bgcolor="#FFFFFF">
<img align='right' src='/adm/lonIcons/lonlogos.gif' />
<h1>Advanced Catalog Search</h1>
<hr />
Enter terms or phrases separated by search operators 
such as AND, OR, or NOT.<br />
<form method="post" action="/adm/searchcat">
$advanced_buttons
$hidden
<table>
<tr><td><font color="#800000" face="helvetica"><b>VIEW:</b></font></td>
<td>
<!-- view selection -->
<select name="viewselect" size ="1">
    <option value="Detailed Citation View" selected="true">Detailed Citation View</option>
    <option value="Summary View">Summary View</option>
    <option value="Fielded Format">Fielded Format</option>
    <option value="XML/SGML">XML/SGML</option>
</select>
<!-- end of view selection -->
</td></tr>
ENDHEADER
    $scrout.=&searchphrasefield('title',   'title'   ,$ENV{'form.title'});
    $scrout.=&searchphrasefield('author',  'author'  ,$ENV{'form.author'});
    $scrout.=&searchphrasefield('subject', 'subject' ,$ENV{'form.subject'});
    $scrout.=&searchphrasefield('keywords','keywords',$ENV{'form.keywords'});
    $scrout.=&searchphrasefield('URL',     'url'     ,$ENV{'form.url'});
    $scrout.=&searchphrasefield('notes',   'notes'   ,$ENV{'form.notes'});
    $scrout.=&searchphrasefield('abstract','abstract',$ENV{'form.abstract'});
    # Hack - an empty table row.
    $scrout.="<tr><td>&nbsp;</td><td>&nbsp;</td></tr>\n";
    $scrout.=&searchphrasefield('file<br />extension','mime',
                        $ENV{'form.mime'});
    $scrout.="<tr><td>&nbsp;</td><td>&nbsp;</td></tr>\n";
    $scrout.=&searchphrasefield('publisher<br />owner','owner',
				$ENV{'form.owner'});
    $scrout.="</table>\n";
    $ENV{'form.category'}='any' unless length($ENV{'form.category'});
    $scrout.=&selectbox('File Category','category',
			$ENV{'form.category'},
			'any','Any category',
			undef,
			(&Apache::loncommon::filecategories()));
    $ENV{'form.language'}='any' unless length($ENV{'form.language'});
    #----------------------------------------------------------------
    # Allow restriction to multiple domains.
    #   I make the crazy assumption that there will never be a domain 'any'.
    #
    $ENV{'form.domains'} = 'any' if (! exists($ENV{'form.domains'}));
    my @allowed_domains = (ref($ENV{'form.domains'}) ? @{$ENV{'form.domains'}} 
                           :  ($ENV{'form.domains'}) );
    my %domain_hash = ();
    foreach (@allowed_domains) {
        $domain_hash{$_}++;
    }
    my @domains =&Apache::loncommon::get_domains();
    # adjust the size of the select box
    my $size = 4;
    my $size = (scalar @domains < ($size - 1) ? scalar @domains + 1 : $size);
    # standalone machines do not get to choose a domain to search.
    if ((scalar @domains) == 1) {
        $scrout .='<input type="hidden" name="domains" value="any" />'."\n";
    } else {
        $scrout.="\n".'<font color="#800000" face="helvetica"><b>'.
            'DOMAINS</b></font><br />'.
                '<select name="domains" size="'.$size.'" multiple>'."\n".
                    '<option name="any" value="any" '.
                        ($domain_hash{'any'}? 'selected ' :'').
                        '>all domains</option>'."\n";
        foreach my $dom (sort @domains) {
            $scrout.="<option name=\"$dom\" ".
                ($domain_hash{$dom} ? 'selected ' :'').">$dom</option>\n";
        }
        $scrout.="</select>\n";
    }
    #----------------------------------------------------------------
    $scrout.=&selectbox('Limit by language','language',
			$ENV{'form.language'},'any','Any Language',
			\&{Apache::loncommon::languagedescription},
			(&Apache::loncommon::languageids),
			);
# ------------------------------------------------ Compute date selection boxes
    $scrout.=<<CREATIONDATESTART;
<p>
<font color="#800000" face="helvetica"><b>LIMIT BY CREATION DATE RANGE:</b>
</font>
<br />
between:
CREATIONDATESTART
    $scrout.=&dateboxes('creationdatestart',1,1,1976,
			$ENV{'form.creationdatestart_month'},
			$ENV{'form.creationdatestart_day'},
			$ENV{'form.creationdatestart_year'},
			);
    $scrout.="and:\n";
    $scrout.=&dateboxes('creationdateend',12,31,2051,
			$ENV{'form.creationdateend_month'},
			$ENV{'form.creationdateend_day'},
			$ENV{'form.creationdateend_year'},
			);
    $scrout.="</p>";
    $scrout.=<<LASTREVISIONDATESTART;
<p>
<font color="#800000" face="helvetica"><b>LIMIT BY LAST REVISION DATE RANGE:
</b></font>
<br />between:
LASTREVISIONDATESTART
    $scrout.=&dateboxes('lastrevisiondatestart',1,1,1976,
			$ENV{'form.lastrevisiondatestart_month'},
			$ENV{'form.lastrevisiondatestart_day'},
			$ENV{'form.lastrevisiondatestart_year'},
			);
    $scrout.=<<LASTREVISIONDATEEND;
and:
LASTREVISIONDATEEND
    $scrout.=&dateboxes('lastrevisiondateend',12,31,2051,
			$ENV{'form.lastrevisiondateend_month'},
			$ENV{'form.lastrevisiondateend_day'},
			$ENV{'form.lastrevisiondateend_year'},
			);
    $scrout.='</p>';
    $ENV{'form.copyright'}='any' unless length($ENV{'form.copyright'});
    $scrout.=&selectbox('Limit by copyright/distribution','copyright',
			 $ENV{'form.copyright'},
			 'any','Any copyright/distribution',
			 \&{Apache::loncommon::copyrightdescription},
			 (&Apache::loncommon::copyrightids),
			 );
# ------------------------------------------- Compute customized metadata field
    $scrout.=<<CUSTOMMETADATA;
<p>
<font color="#800000" face="helvetica"><b>LIMIT BY SPECIAL METADATA FIELDS:</b>
</font>
For resource-specific metadata, enter in an expression in the form of 
<i>key</i>=<i>value</i> separated by operators such as AND, OR or NOT.<br />
<b>Example:</b> grandmother=75 OR grandfather=85
<br />
CUSTOMMETADATA
    $scrout.=&simpletextfield('custommetadata',$ENV{'form.custommetadata'});
    $scrout.=<<CUSTOMSHOW;
<p>
<font color="#800000" face="helvetica"><b>SHOW SPECIAL METADATA FIELDS:</b>
</font>
Enter in a space-separated list of special metadata fields to show
in a fielded listing for each record result.
<br />
CUSTOMSHOW
    $scrout.=&simpletextfield('customshow',$ENV{'form.customshow'});
    $scrout.=<<ENDDOCUMENT;
$advanced_buttons
</form>
</body>
</html>
ENDDOCUMENT
    return $scrout;
}

######################################################################
######################################################################

=pod 

=item &make_persistent() 

Returns a scalar which holds the current ENV{'form.*'} values in
a 'hidden' html input tag.  This allows search interface information
to be somewhat persistent.

=cut

######################################################################
######################################################################

sub make_persistent {
    my %save = %{shift()};
    my $persistent='';
    foreach (keys %save) {
	if (/^form\./ && !/submit/) {
	    my $name=$_;
            my @values = (ref($save{$name}) ? @{$save{$name}} : ($save{$name}));
	    $name=~s/^form\.//;
            foreach (@values) {
                s/\"/\'/g; # do not mess with html field syntax
                $persistent.=<<END;
<input type="hidden" name="$name" value="$_" />
END
            }
        }
    }
    return $persistent;
}


######################################################################
######################################################################

=pod 

=item HTML form building functions

=over 4

=item &simpletextfield() 

Inputs: $name,$value,$size

Returns a text input field with the given name, value, and size.  
If size is not specified, a value of 20 is used.

=item &simplecheckbox()

Inputs: $name,$value

Returns a simple check box with the given $name.
If $value eq 'on' the box is checked.

=item &searchphrasefield()

Inputs: $title,$name,$value

Returns html for a title line and an input field for entering search terms.
the instructions "Enter terms or phrases separated by search operators such 
as AND, OR, or NOT." are given following the title.  The entry field (which
is where the $name and $value are used) is an 80 column simpletextfield.

=item &dateboxes()

Returns html selection form elements for the specification of 
the day, month, and year.

=item &selectbox()

Returns a scalar containing an html <select> form.  

Inputs: 

=over 4

=item $title 

Printed above the select box, in uppercase.

=item $name 

The name element of the <select> tag.

=item $default 

The default value of the form.  Can be $anyvalue or in @idlist.

=item $anyvalue 

The <option value="..."> used to indicate a default of 
none of the values.

=item $anytag 

The text associate with $anyvalue above.

=item $functionref 

Each element in @idlist will be passed as a parameter 
to the function referenced here.  The return value of the function should
be a scalar description of the items.  If this value is undefined the 
description of each item in @idlist will be the item name.

=item @idlist 

The items to be selected from.  One of these or $anyvalue will be the 
value returned by the form element, $ENV{form.$name}.

=back

=back 

=cut

######################################################################
######################################################################

sub simpletextfield {
    my ($name,$value,$size)=@_;
    $size = 20 if (! defined($size));
    return '<input type="text" name="'.$name.
        '" size="'.$size.'" value="'.$value.'" />';
}

sub simplecheckbox {
    my ($name,$value)=@_;
    my $checked='';
    $checked="checked" if $value eq 'on';
    return '<input type="checkbox" name="'.$name.'" '. $checked . ' />';
}

sub searchphrasefield {
    my ($title,$name,$value)=@_;
    my $uctitle=uc($title);
    return '<tr><td><font color="#800000" face="helvetica">'.
        '<b>'.$uctitle.':&nbsp;</b></font></td><td>'.
                &simpletextfield($name,$value,50)."</td></tr>\n";
}

sub dateboxes {
    my ($name,$defaultmonth,$defaultday,$defaultyear,
	$currentmonth,$currentday,$currentyear)=@_;
    ($defaultmonth,$defaultday,$defaultyear)=('','','');
    #
    # Day
    my $day=<<END;
<select name="${name}_day">
<option value='$defaultday'> </option>
END
    for (my $i = 1; $i<=31; $i++) {
	$day.="<option value=\"$i\">$i</option>\n";
    }
    $day.="</select>\n";
    $day=~s/(\"$currentday\")/$1 SELECTED/ if length($currentday);
    #
    # Month
    my $month=<<END;
<select name="${name}_month">
<option value='$defaultmonth'> </option>
END
    my $i = 1;
    foreach (qw/January February March April May June 
	     July August September October November December /){
	$month .="<option value=\"$i\">$_</option>\n";
	$i++;
    }
    $month.="</select>\n";
    $month=~s/(\"$currentmonth\")/$1 SELECTED/ if length($currentmonth);
    #
    # Year (obviously)
    my $year=<<END;
<select name="${name}_year">
<option value='$defaultyear'> </option>
END
    my $maxyear = 2051; 
    for (my $i = 1976; $i<=$maxyear; $i++) {
	$year.="<option value=\"$i\">$i</option>\n";
    }
    $year.="</select>\n";
    $year=~s/(\"$currentyear\")/$1 SELECTED/ if length($currentyear);
    return "$month$day$year";
}

sub selectbox {
    my ($title,$name,$default,$anyvalue,$anytag,$functionref,@idlist)=@_;
    if (! defined($functionref)) { $functionref = sub { $_[0]}; }
    my $uctitle=uc($title);
    my $selout="\n".'<p><font color="#800000" face="helvetica">'.
        '<b>'.$uctitle.': </b></font><select name="'.$name.'">';
    foreach ($anyvalue,@idlist) {
        $selout.='<option value="'.$_.'"';
        if ($_ eq $default and !/^any$/) {
	    $selout.=' selected >'.&{$functionref}($_).'</option>';
	}
	elsif ($_ eq $default and /^$anyvalue$/) {
	    $selout.=' selected >'.$anytag.'</option>';
	}
        else {$selout.='>'.&{$functionref}($_).'</option>';}
    }
    return $selout.'</select></p>';
}

######################################################################
######################################################################

=pod 

=item &parse_advanced_search()

Parse advanced search form and return the following:

=over 4

=item $query Scalar containing an SQL query.

=item $customquery Scalar containing a custom query.

=item $customshow Scalar containing commands to show custom metadata.

=item $libraries_to_query Reference to array of domains to search.

=back

=cut

######################################################################
######################################################################
sub parse_advanced_search {
    my ($r)=@_;
    my $fillflag=0;
    # Clean up fields for safety
    for my $field ('title','author','subject','keywords','url','version',
		   'creationdatestart_month','creationdatestart_day',
		   'creationdatestart_year','creationdateend_month',
		   'creationdateend_day','creationdateend_year',
		   'lastrevisiondatestart_month','lastrevisiondatestart_day',
		   'lastrevisiondatestart_year','lastrevisiondateend_month',
		   'lastrevisiondateend_day','lastrevisiondateend_year',
		   'notes','abstract','mime','language','owner',
		   'custommetadata','customshow','category') {
	$ENV{"form.$field"}=~s/[^\w\/\s\(\)\=\-\"\']//g;
    }
    foreach ('mode','form','element') {
	# is this required?  Hmmm.
	next unless (exists($ENV{"form.$_"}));
	$ENV{"form.$_"}=&Apache::lonnet::unescape($ENV{"form.$_"});
	$ENV{"form.$_"}=~s/[^\w\/\s\(\)\=\-\"\']//g;
    }
    # Preprocess the category form element.
    if ($ENV{'form.category'} ne 'any') {
        my @extensions = &Apache::loncommon::filecategorytypes
            ($ENV{'form.category'});
        $ENV{'form.mime'} = join ' OR ',@extensions;
    }
    # Check to see if enough information was filled in
    for my $field ('title','author','subject','keywords','url','version',
		   'notes','abstract','mime','language','owner',
		   'custommetadata') {
	if (&filled($ENV{"form.$field"})) {
	    $fillflag++;
	}
    }
    unless ($fillflag) {
	&output_blank_field_error($r);
	return ;
    }
    # Turn the form input into a SQL-based query
    my $query='';
    my @queries;
    # Evaluate logical expression AND/OR/NOT phrase fields.
    foreach my $field ('title','author','subject','notes','abstract','url',
		       'keywords','version','owner','mime') {
	if ($ENV{'form.'.$field}) {
	    push @queries,&build_SQL_query($field,$ENV{'form.'.$field});
        }
    }
    # I dislike the hack below.
    if ($ENV{'form.category'}) {
        $ENV{'form.mime'}='';
    }
    # Evaluate option lists
    if ($ENV{'form.language'} and $ENV{'form.language'} ne 'any') {
	push @queries,"(language like \"$ENV{'form.language'}\")";
    }
    if ($ENV{'form.copyright'} and $ENV{'form.copyright'} ne 'any') {
	push @queries,"(copyright like \"$ENV{'form.copyright'}\")";
    }
    # Evaluate date windows
    my $datequery=&build_date_queries(
			$ENV{'form.creationdatestart_month'},
			$ENV{'form.creationdatestart_day'},
			$ENV{'form.creationdatestart_year'},
			$ENV{'form.creationdateend_month'},
			$ENV{'form.creationdateend_day'},
			$ENV{'form.creationdateend_year'},
			$ENV{'form.lastrevisiondatestart_month'},
			$ENV{'form.lastrevisiondatestart_day'},
			$ENV{'form.lastrevisiondatestart_year'},
			$ENV{'form.lastrevisiondateend_month'},
			$ENV{'form.lastrevisiondateend_day'},
			$ENV{'form.lastrevisiondateend_year'},
			);
    # Test to see if date windows are legitimate
    if ($datequery=~/^Incorrect/) {
	&output_date_error($r,$datequery);
	return ;
    }
    elsif ($datequery) {
	push @queries,$datequery;
    }
    # Process form information for custom metadata querying
    my $customquery=undef;
    if ($ENV{'form.custommetadata'}) {
	$customquery=&build_custommetadata_query('custommetadata',
				      $ENV{'form.custommetadata'});
    }
    my $customshow=undef;
    if ($ENV{'form.customshow'}) {
	$customshow=$ENV{'form.customshow'};
	$customshow=~s/[^\w\s]//g;
	my @fields=split(/\s+/,$customshow);
	$customshow=join(" ",@fields);
    }
    ## ---------------------------------------------------------------
    ## Deal with restrictions to given domains
    ## 
    my $libraries_to_query = undef;
    # $ENV{'form.domains'} can be either a scalar or an array reference.
    # We need an array.
    my @allowed_domains = (ref($ENV{'form.domains'}) ? @{$ENV{'form.domains'}} 
                           :  ($ENV{'form.domains'}) );
    my %domain_hash = ();
    foreach (@allowed_domains) {
        $domain_hash{$_}++;
    }
    foreach (keys(%Apache::lonnet::libserv)) {
        if ($_ eq 'any') {
            $libraries_to_query = undef;
            last;
        }
        if (exists($domain_hash{$Apache::lonnet::hostdom{$_}})) {
            push @$libraries_to_query,$_;
        }
    }
    #
    if (@queries) {
	$query=join(" AND ",@queries);
	$query="select * from metadata where $query";
    } elsif ($customquery) {
        $query = '';
    }
    return ($query,$customquery,$customshow,$libraries_to_query);
}

######################################################################
######################################################################

=pod 

=item &parse_basic_search() 

Parse the basic search form and return a scalar containing an sql query.

=cut

######################################################################
######################################################################
sub parse_basic_search {
    my ($r)=@_;
    # Clean up fields for safety
    for my $field ('basicexp') {
	$ENV{"form.$field"}=~s/[^\w\s\(\)\-]//g;
    }
    foreach ('mode','form','element') {
	# is this required?  Hmmm.
	next unless (exists($ENV{"form.$_"}));
	$ENV{"form.$_"}=&Apache::lonnet::unescape($ENV{"form.$_"});
	$ENV{"form.$_"}=~s/[^\w\/\s\(\)\=\-\"\']//g;
    }

    # Check to see if enough is filled in
    unless (&filled($ENV{'form.basicexp'})) {
	&output_blank_field_error($r);
	return OK;
    }

    # Build SQL query string based on form page
    my $query='';
    my $concatarg=join(',"    ",',
		       ('title', 'author', 'subject', 'notes', 'abstract',
                        'keywords'));
    $concatarg='title' if $ENV{'form.titleonly'};

    $query=&build_SQL_query('concat('.$concatarg.')',$ENV{'form.'.'basicexp'});
    return 'select * from metadata where '.$query;
}


######################################################################
######################################################################

=pod 

=item &build_SQL_query() 

Builds a SQL query string from a logical expression with AND/OR keywords
using Text::Query and &recursive_SQL_query_builder()

=cut

######################################################################
######################################################################
sub build_SQL_query {
    my ($field_name,$logic_statement)=@_;
    my $q=new Text::Query('abc',
			  -parse => 'Text::Query::ParseAdvanced',
			  -build => 'Text::Query::Build');
    $q->prepare($logic_statement);
    my $matchexp=${$q}{'matchexp'}; chomp $matchexp;
    my $sql_query=&recursive_SQL_query_build($field_name,$matchexp);
    return $sql_query;
}

######################################################################
######################################################################

=pod 

=item &build_custommetadata_query() 

Constructs a custom metadata query using a rather heinous regular
expression.

=cut

######################################################################
######################################################################
sub build_custommetadata_query {
    my ($field_name,$logic_statement)=@_;
    my $q=new Text::Query('abc',
			  -parse => 'Text::Query::ParseAdvanced',
			  -build => 'Text::Query::BuildAdvancedString');
    $q->prepare($logic_statement);
    my $matchexp=${$q}{'-parse'}{'-build'}{'matchstring'};
    # quick fix to change literal into xml tag-matching
    # will eventually have to write a separate builder module
    # wordone=wordtwo becomes\<wordone\>[^\<] *wordtwo[^\<]*\<\/wordone\>
    $matchexp =~ s/(\w+)\\=([\w\\\+]+)?# wordone=wordtwo is changed to 
                 /\\<$1\\>?#           \<wordone\>
                   \[\^\\<\]?#        [^\<]         
                   \*$2\[\^\\<\]?#           *wordtwo[^\<]
                   \*\\<\\\/$1\\>?#                        *\<\/wordone\>
                   /g;
    return $matchexp;
}

######################################################################
######################################################################

=pod 

=item &recursive_SQL_query_build() 

Recursively constructs an SQL query.  Takes as input $dkey and $pattern.

=cut

######################################################################
######################################################################
sub recursive_SQL_query_build {
    my ($dkey,$pattern)=@_;
    my @matches=($pattern=~/(\[[^\]|\[]*\])/g);
    return $pattern unless @matches;
    foreach my $match (@matches) {
	$match=~/\[ (\w+)\s(.*) \]/;
	my ($key,$value)=($1,$2);
	my $replacement='';
	if ($key eq 'literal') {
	    $replacement="($dkey like \"\%$value\%\")";
	}
	elsif ($key eq 'not') {
	    $value=~s/like/not like/;
#	    $replacement="($dkey not like $value)";
	    $replacement="$value";
	}
	elsif ($key eq 'and') {
	    $value=~/(.*[\"|\)]) ([|\(|\^].*)/;
	    $replacement="($1 AND $2)";
	}
	elsif ($key eq 'or') {
	    $value=~/(.*[\"|\)]) ([|\(|\^].*)/;
	    $replacement="($1 OR $2)";
	}
	substr($pattern,
	       index($pattern,$match),
	       length($match),
	       $replacement
	       );
    }
    &recursive_SQL_query_build($dkey,$pattern);
}

######################################################################
######################################################################

=pod 

=item &build_date_queries() 

Builds a SQL logic query to check time/date entries.
Also reports errors (check for /^Incorrect/).

=cut

######################################################################
######################################################################
sub build_date_queries {
    my ($cmonth1,$cday1,$cyear1,$cmonth2,$cday2,$cyear2,
	$lmonth1,$lday1,$lyear1,$lmonth2,$lday2,$lyear2)=@_;
    my @queries;
    if ($cmonth1 or $cday1 or $cyear1 or $cmonth2 or $cday2 or $cyear2) {
	unless ($cmonth1 and $cday1 and $cyear1 and
		$cmonth2 and $cday2 and $cyear2) {
	    return "Incorrect entry for the creation date.  You must specify ".
		   "a starting month, day, and year and an ending month, ".
		   "day, and year.";
	}
	my $cnumeric1=sprintf("%d%2d%2d",$cyear1,$cmonth1,$cday1);
	$cnumeric1+=0;
	my $cnumeric2=sprintf("%d%2d%2d",$cyear2,$cmonth2,$cday2);
	$cnumeric2+=0;
	if ($cnumeric1>$cnumeric2) {
	    return "Incorrect entry for the creation date.  The starting ".
		   "date must occur before the ending date.";
	}
	my $cquery="(creationdate BETWEEN '$cyear1-$cmonth1-$cday1' AND '".
	           "$cyear2-$cmonth2-$cday2 23:59:59')";
	push @queries,$cquery;
    }
    if ($lmonth1 or $lday1 or $lyear1 or $lmonth2 or $lday2 or $lyear2) {
	unless ($lmonth1 and $lday1 and $lyear1 and
		$lmonth2 and $lday2 and $lyear2) {
	    return "Incorrect entry for the last revision date.  You must ".
		   "specify a starting month, day, and year and an ending ".
		   "month, day, and year.";
	}
	my $lnumeric1=sprintf("%d%2d%2d",$lyear1,$lmonth1,$lday1);
	$lnumeric1+=0;
	my $lnumeric2=sprintf("%d%2d%2d",$lyear2,$lmonth2,$lday2);
	$lnumeric2+=0;
	if ($lnumeric1>$lnumeric2) {
	    return "Incorrect entry for the last revision date.  The ".
		   "starting date must occur before the ending date.";
	}
	my $lquery="(lastrevisiondate BETWEEN '$lyear1-$lmonth1-$lday1' AND '".
	           "$lyear2-$lmonth2-$lday2 23:59:59')";
	push @queries,$lquery;
    }
    if (@queries) {
	return join(" AND ",@queries);
    }
    return '';
}

######################################################################
######################################################################

=pod 

=item &output_results() 

Format and output results based on a reply list.
There are two windows that this function writes to.  The main search
window ("srch") has a listing of the results.  A secondary window ("popwin")
gives the status of the network search (time elapsed, number of machines
contacted, etc.)

=cut

######################################################################
######################################################################
sub output_results {
#    &Apache::lonnet::logthis("output_results:".time);
    my $fnum; # search result counter
    my ($mode,$r,$replyref,$hidden)=@_;
    my %rhash=%{$replyref};
    my $compiledresult='';
    my $timeremain=300; # (seconds)
    my $elapsetime=0;
    my $resultflag=0;
    my $tflag=1;
    my $viewselect=$ENV{'form.viewselect'};
    #
    # make query information persistent to allow for subsequent revision
    my $persistent=&make_persistent(\%ENV);
    # spit out the results header
    $r->print(&search_results_header($mode));
    $r->rflush();
    # begin showing the cataloged results
    my $action = "/adm/searchcat";
    if ($mode eq 'Basic') { 
        $action .= "?reqinterface=basic";
    } elsif ($mode eq 'Advanced') {
        $action .= "?reqinterface=advanced";
    }
    $r->print(<<CATALOGCONTROLS);
<form name='results' method="post" action="$action">
$hidden
<input type='hidden' name='acts' value='' />
<input type='button' value='Revise search request'
onClick='this.form.submit();' />
$importbutton
$closebutton
$persistent
<hr />
CATALOGCONTROLS
    #
    # make the pop-up window for status
    #
    $r->print(&make_popwin(%rhash));
    $r->rflush();
    ##
    ## Prepare for the main loop below
    ##
    my $servercount=0;
    my $hitcountsum=0;
    my $servernum=(keys %rhash);
    my $serversleft=$servernum;
    ##
    ## Run until we run out of time or we run out of servers
    ##
    while($serversleft && $timeremain) {
      ##
      ## %rhash has servers deleted from it as results come in 
      ## (within the foreach loop below).
      ##
      foreach my $rkey (sort keys %rhash) {
#        &Apache::lonnet::logthis("Server $rkey:".time);
	$servercount++;
	$compiledresult='';
	my $reply=$rhash{$rkey};
	my @results;
	if ($reply eq 'con_lost') {
	    &popwin_imgupdate($r,$rkey,"srvbad.gif");
	    $serversleft--;
            delete $rhash{$rkey};
	} else {
            # must do since 'use strict' checks for tainting
	    $reply=~/^([\.\w]+)$/; 
	    my $replyfile=$r->dir_config('lonDaemons').'/tmp/'.$1;
	    $reply=~/(.*?)\_/;
            for (my $counter=0;$counter<2;$counter++) {
                if (-e $replyfile && ! -e "$replyfile.end") {
                    &popwin_imgupdate($r,$rkey,"srvhalf.gif");
                    &popwin_js($r,'popwin.hc["'.$rkey.'"]='.
                               '"still transferring..."'.';');
                }
                # Are we finished transferring data?
                if (-e "$replyfile.end") {
                    $serversleft--;
                    delete $rhash{$rkey};
                    if (-s $replyfile) {
                        &popwin_imgupdate($r,$rkey,"srvgood.gif");
                        my $fh;
                        unless ($fh=Apache::File->new($replyfile)){ 
                            # Is it really appropriate to die on this error?
                            $r->print('ERROR: file '.
                                      $replyfile.' cannot be opened');
                            return OK;
                        }
                        @results=<$fh> if $fh;
                        my $hits =@results;
                        &popwin_js($r,'popwin.hc["'.$rkey.'"]='.
                                   $hits.';');
                        $hitcountsum+=$hits;
                        &popwin_js($r,'popwin.document.forms.popremain.'.
                                   'numhits.value='.$hitcountsum.';');
                    } else {
                        &popwin_imgupdate($r,$rkey,"srvempty.gif");
                        &popwin_js($r,'popwin.hc["'.$rkey.'"]=0;');
                    }
                    last;
                } # end of if ( -e "$replyfile.end")
                last unless $timeremain;
                sleep 1;    # wait for daemons to write files?
                $timeremain--;
                $elapsetime++;
                &popwin_js($r,"popwin.document.popremain.".
                           "elapsetime.value=$elapsetime;");
	    }
	    &popwin_js($r,'popwin.document.whirly.'.
		       'src="/adm/lonIcons/lonanimend.gif";');
	} # end of if ($reply eq 'con_lost') else statement
        my %Fields = undef;     # Holds the data to be sent to the various 
                                # *_view routines.
        my ($extrashow,$customfields,$customhash) = &handle_custom_fields(\@results);
        my @customfields = @$customfields;
        my %customhash   = %$customhash;
	untie %hash if (keys %hash);
        #
	if (! tie(%hash,'GDBM_File',$diropendb,&GDBM_WRCREAT,0640)) {
	    $r->print('<html><head></head><body>Unable to tie hash to db '.
                      'file</body></html>');
        } else {
	    if ($ENV{'form.launch'} eq '1') {
		&start_fresh_session();
	    }
	    foreach my $result (@results) {
		next if $result=~/^custom\=/;
		chomp $result;
		next unless $result;
                %Fields = &parse_raw_result($result,$rkey);
		$Fields{'extrashow'}=$extrashow;
		if ($extrashow) {
		    foreach my $field (@customfields) {
			my $value='';
			$value = $1 if ($customhash{$Fields{'url'}}=~/\<{$field}[^\>]*\>(.*?)\<\/{$field}[^\>]*\>/s);
                        $Fields{'extrashow'}=~s/\<\!\-\- $field \-\-\>/ $value/g;
                    }
                }
                $compiledresult.="\n<p>\n";
                if ($ENV{'form.catalogmode'} eq 'interactive') {
                    my $titleesc=$Fields{'title'};
                    $titleesc=~s/\'/\\'/; # '
                    $compiledresult.=<<END if ($ENV{'form.catalogmode'} eq 'interactive');
<font size='-1'><INPUT TYPE="button" NAME="returnvalues" VALUE="SELECT"
onClick="javascript:select_data('$titleesc','$Fields{'url'}')">
</font>
<br />
END
                }
                if ($ENV{'form.catalogmode'} eq 'groupsearch') {
		    $fnum+=0;
		    $hash{"pre_${fnum}_link"}=$Fields{'url'};
		    $hash{"pre_${fnum}_title"}=$Fields{'title'};
		    $compiledresult.=<<END;
<font size='-1'>
<input type="checkbox" name="returnvalues" value="SELECT"
onClick="javascript:queue($fnum)" />
</font>
<br />
END
# <input type="hidden" name="title$fnum" value="$title" />
# <input type="hidden" name="url$fnum" value="$url" />
                    $fnum++;
		}
	        if ($viewselect eq 'Detailed Citation View') {
		    $compiledresult.=&detailed_citation_view
                        (%Fields, hostname => $rkey );
		}
                elsif ($viewselect eq 'Summary View') {
		    $compiledresult.=&summary_view
                        (%Fields, hostname => $rkey );
	        }
                elsif ($viewselect eq 'Fielded Format') {
		    $compiledresult.=&fielded_format_view
                        (%Fields, hostname => $rkey );
	        }
                elsif ($viewselect eq 'XML/SGML') {
		    $compiledresult.=&xml_sgml_view
                        (%Fields, hostname => $rkey );
		}
                if ($compiledresult or $servercount!=$servernum) {
                    $compiledresult.="<hr align='left' width='200' noshade />";
                }
            }
            untie %hash;
        }
	if ($compiledresult) {
	    $resultflag=1;
            $r->print($compiledresult);
	}
        my $percent=sprintf('%3.0f',($servercount/$servernum*100));
      } # End of foreach loop over servers remaining
    }   # End of big loop - while($serversleft && $timeremain)
    unless ($resultflag) {
        $r->print("\nThere were no results that matched your query\n");
    }
    $r->print('<script type="text/javascript">'.'popwin.close()</script>'.
              "\n"); 
    $r->print("</body>\n</html>\n");
    $r->rflush(); 
    return;
}

###########################################################
###########################################################

=pod

=item &parse_raw_result()

Takes a line from the file of results and parse it.  Returns a hash 
with keys for the following fields:
'title', 'author', 'subject', 'url', 'keywords', 'version', 'notes', 
'abstract', 'mime', 'lang', 'owner', 'copyright', 'creationdate', 
'lastrevisiondate'.

In addition, the following tags are set by calling the appropriate 
lonnet function: 'language', 'cprtag', 'mimetag'.

The 'title' field is set to "Untitled" if the title field is blank.

'abstract' and 'keywords' are truncated to 200 characters.

=cut

###########################################################
###########################################################
sub parse_raw_result {
    my ($result,$hostname) = @_;
    # Check for a comma - if it is there then we do not need to unescape the
    # string.  There seems to be some kind of problem with some items in
    # the database - the entire string gets sent out unescaped...?
    unless ($result =~ /,/) {
        $result = &Apache::lonnet::unescape($result);
    }
    my @fields=map {
        &Apache::lonnet::unescape($_);
    } (split(/\,/,$result));
    my ($title,$author,$subject,$url,$keywords,$version,
        $notes,$abstract,$mime,$lang,
        $creationdate,$lastrevisiondate,$owner,$copyright)=@fields;
    my %Fields = 
        ( title     => &Apache::lonnet::unescape($title),
          author    => &Apache::lonnet::unescape($author),
          subject   => &Apache::lonnet::unescape($subject),
          url       => &Apache::lonnet::unescape($url),
          keywords  => &Apache::lonnet::unescape($keywords),
          version   => &Apache::lonnet::unescape($version),
          notes     => &Apache::lonnet::unescape($notes),
          abstract  => &Apache::lonnet::unescape($abstract),
          mime      => &Apache::lonnet::unescape($mime),
          lang      => &Apache::lonnet::unescape($lang),
          owner     => &Apache::lonnet::unescape($owner),
          copyright => &Apache::lonnet::unescape($copyright),
          creationdate     => &Apache::lonnet::unescape($creationdate),
          lastrevisiondate => &Apache::lonnet::unescape($lastrevisiondate)
        );
    $Fields{'language'} = 
        &Apache::loncommon::languagedescription($Fields{'lang'});
    $Fields{'copyrighttag'} =
        &Apache::loncommon::copyrightdescription($Fields{'copyright'});
    $Fields{'mimetag'} =
        &Apache::loncommon::filedescription($Fields{'mime'});
    if ($Fields{'author'}=~/^(\s*|error)$/) {
        $Fields{'author'}="Unknown Author";
    }
    # Put spaces in the keyword list, if needed.
    $Fields{'keywords'}=~ s/,([A-z])/, $1/g; 
    if ($Fields{'title'}=~ /^\s*$/ ) { 
        $Fields{'title'}='Untitled'; 
    }
    unless ($ENV{'user.adv'}) {
        $Fields{'keywords'} = '- not displayed -';
        $Fields{'notes'}    = '- not displayed -';
        $Fields{'abstract'} = '- not displayed -';
        $Fields{'subject'}  = '- not displayed -';
    }
    if (length($Fields{'abstract'})>200) {
        $Fields{'abstract'} = 
            substr($Fields{'abstract'},0,200).'...';
    }
    if (length($Fields{'keywords'})>200) {
        $Fields{'keywords'} =
            substr($Fields{'keywords'},0,200).'...';
    }
    return %Fields;
}

###########################################################
###########################################################

=pod

=item &handle_custom_fields()

=cut

###########################################################
###########################################################
sub handle_custom_fields {
    my @results = @{shift()};
    my $customshow='';
    my $extrashow='';
    my @customfields;
    if ($ENV{'form.customshow'}) {
        $customshow=$ENV{'form.customshow'};
        $customshow=~s/[^\w\s]//g;
        my @fields=map {
            "<font color=\"#008000\">$_:</font><!-- $_ -->";
        } split(/\s+/,$customshow);
        @customfields=split(/\s+/,$customshow);
        if ($customshow) {
            $extrashow="<ul><li>".join("</li><li>",@fields)."</li></ul>\n";
        }
    }
    my $customdata='';
    my %customhash;
    foreach my $result (@results) {
        if ($result=~/^(custom\=.*)$/) { # grab all custom metadata
            my $tmp=$result;
            $tmp=~s/^custom\=//;
            my ($k,$v)=map {&Apache::lonnet::unescape($_);
                        } split(/\,/,$tmp);
            $customhash{$k}=$v;
        }
    }
    return ($extrashow,\@customfields,\%customhash);
}

######################################################################
######################################################################

=pod

=item &search_results_header

Output the proper html headers and javascript code to deal with different 
calling modes.

Takes most inputs directly from %ENV, except $mode.  

=over 4

=item $mode is either (at this writing) 'Basic' or 'Advanced'

=back

The following environment variables are checked:

=over 4

=item 'form.catalogmode' 

Checked for 'interactive' and 'groupsearch'.

=item 'form.mode'

Checked for existance & 'edit' mode.

=item 'form.form'

=item 'form.element'

=back

=cut

######################################################################
######################################################################
sub search_results_header {
    my ($mode) = @_;
    $mode = lc($mode);
    my $title;
    if ($mode eq 'advanced') {
        $title = "Advanced Search Results";
    } elsif ($mode eq 'basic') {
        $title = "Basic Search Results";
    }
    my $result = '';
    # output beginning of search page
    $result.=<<BEGINNING;
<html>
<head>
<title>$title</title>
BEGINNING
    # conditional output of script functions dependent on the mode in
    # which the search was invoked
    if ($ENV{'form.catalogmode'} eq 'interactive'){
	if (! exists($ENV{'form.mode'}) || $ENV{'form.mode'} ne 'edit') {
            $result.=<<SCRIPT;
<script type="text/javascript">
    function select_data(title,url) {
	changeTitle(title);
	changeURL(url);
	self.close();
    }
    function changeTitle(val) {
	if (opener.inf.document.forms.resinfo.elements.t) {
	    opener.inf.document.forms.resinfo.elements.t.value=val;
	}
    }
    function changeURL(val) {
	if (opener.inf.document.forms.resinfo.elements.u) {
	    opener.inf.document.forms.resinfo.elements.u.value=val;
	}
    }
</script>
SCRIPT
        } elsif ($ENV{'form.mode'} eq 'edit') {
            my $form = $ENV{'form.form'};
            my $element = $ENV{'form.element'};
            $result.=<<SCRIPT;
<script type="text/javascript">
function select_data(title,url) {
    changeURL(url);
    self.close();
}
function changeTitle(val) {
}
function changeURL(val) {
    if (window.opener.document) {
        window.opener.document.forms["$form"].elements["$element"].value=val;
    } else {
	var url = 'forms[\"$form\"].elements[\"$element\"].value';
        alert("Unable to transfer data to "+url);
    }
}
</script>
SCRIPT
        }
    }
    $result.=<<SCRIPT if $ENV{'form.catalogmode'} eq 'groupsearch';
<script type="text/javascript">
    function select_data(title,url) {
//	alert('DEBUG: Should be storing '+title+' and '+url);
    }
    function queue(val) {
	if (eval("document.forms.results.returnvalues["+val+"].checked")) {
	    document.forms.results.acts.value+='1a'+val+'b';
	}
	else {
	    document.forms.results.acts.value+='0a'+val+'b';
	}
    }
    function select_group() {
	window.location=
    "/adm/groupsort?mode=$ENV{'form.mode'}&catalogmode=groupsearch&acts="+
	    document.forms.results.acts.value;
    }
</script>
SCRIPT
    $result.=<<SCRIPT;
<script type="text/javascript">
    function displayinfo(val) {
	popwin.document.forms.popremain.sdetails.value=val;
    }
    function openhelp(val) {
	openhelpwin=open('/adm/help/searchcat.html','helpscreen',
	     'scrollbars=1,width=400,height=300');
	openhelpwin.focus();
    }
    function abortsearch(val) {
	popwin.close();
    }
</script>
SCRIPT
    $result.=<<END;
</head>
<body bgcolor="#ffffff">
<img align=right src=/adm/lonIcons/lonlogos.gif>
<h1>$title</h1>
END
    return $result;
}

######################################################################
######################################################################

=pod

=item &make_popwin()

Returns html with javascript in it to open up the status window.

=cut

######################################################################
######################################################################
sub make_popwin {
    my %rhash = @_;
    my $servernum=(keys %rhash);
    my $hcinit;
    my $grid="'<br />'+\n";
    # $sn is the server number, used ONLY to make sure we have
    # rows of 10 each.  No longer used to index images.
    my $sn=1;
    foreach my $sk (sort keys %rhash) {
	$grid.="'<a href=\"";
	$grid.="javascript:opener.displayinfo('+";
	$grid.="\"'\"+'";
	$grid.=$sk;
	my $hc;
	if ($rhash{$sk} eq 'con_lost') {
	    $hc="BAD CONNECTION ";
	}
	else {
	    $hc="'+\"'\"+\"+hc['$sk']+\"+\"'\"+'";
	    $hcinit.="hc[\"$sk\"]=\"not yet connected...\";";
	}
	$grid.=" hitcount=".$hc;
	$grid.=" domain=".$Apache::lonnet::hostdom{$sk};
	$grid.=" IP=".$Apache::lonnet::hostip{$sk};
	# '+"'"+'">'+
	$grid.="'+\"'\"+')\">'+";
	$grid.="\n";
	$grid.="'<img border=\"0\" name=\"img_".$Apache::lonnet::hostdom{$sk}.
            '_'.$sk."\" src=\"/adm/lonIcons/srvnull.gif\" alt=\"".$sk.
                "\" /></a>'+\n";
	$grid.="'<br />'+\n" unless $sn%10;
        $sn++;
    }
    my $result.=<<ENDPOP;
<script type="text/javascript">
    popwin=open('','popwin','scrollbars=1,width=400,height=220');
    popwin.focus();
    popwin.document.writeln('<'+'html>');
    popwin.document.writeln('<'+'head>');
    popwin.document.writeln('<'+'script>');
    popwin.document.writeln('hc=new Array();$hcinit');
    popwin.document.writeln('<'+'/script>');
    popwin.document.writeln('<'+'/head>'+
        '<'+'body bgcolor="#FFFFFF">'+
	'<'+'image name="whirly" align="right" src="/adm/lonIcons/'+
	'lonanim.gif" '+
	'alt="animated logo" />'+
	'<'+'h3>Search Results Progress<'+'/h3>'+
        '<'+'form name="popremain">'+
        '<'+'tt>'+
	'<'+'br clear="all"/><i>PLEASE BE PATIENT</i>'+
	'<'+'br />SCANNING $servernum SERVERS'+
	'<'+'br clear="all" />Number of record hits found '+
	'<'+'input type="text" size="10" name="numhits"'+
	' value="0" />'+
	'<'+'br clear="all" />Time elapsed '+
	'<'+'input type="text" size="10" name="elapsetime"'+
	' value="0" />'+
	'<'+'br />'+
	'SERVER GRID (click on any cell for details)'+
        $grid
        '<'+'br />'+
	'Server details '+
	'<'+'input type="text" size="35" name="sdetails"'+
	' value="" />'+
	'<'+'br />'+
	' <'+'input type="button" name="button"'+
	' value="close this window" '+
	' onClick="javascript:opener.abortsearch()" />'+
	' <'+'input type="button" name="button"'+
	' value="help" onClick="javascript:opener.openhelp()" />'+
	'<'+'/tt>'+
        '<'+'/form>'+
        '<'+'/body><'+'/html>');
    popwin.document.close();
</script>
ENDPOP
    return $result;
}

######################################################################
######################################################################

=pod 

=item Metadata Viewing Functions

Output is a HTML-ified string.
Input arguments are title, author, subject, url, keywords, version,
notes, short abstract, mime, language, creation date,
last revision date, owner, copyright, hostname, and
extra custom metadata to show.

=over 4

=item &detailed_citation_view() 

=cut

######################################################################
######################################################################
sub detailed_citation_view {
    my %values = @_;
    my $result=<<END;
<h3><a href="http://$ENV{'HTTP_HOST'}$values{'url'}" 
    target='search_preview'>$values{'title'}</a></h3>
<p>
<b>$values{'author'}</b>, <i>$values{'owner'}</i><br />

<b>Subject:       </b> $values{'subject'}<br />
<b>Keyword(s):    </b> $values{'keywords'}<br />
<b>Notes:         </b> $values{'notes'}<br />
<b>MIME Type:     </b> $values{'mimetag'}<br />
<b>Language:      </b> $values{'language'}<br />
<b>Copyright/Distribution:</b> $values{'cprtag'}<br />
</p>
$values{'extrashow'}
<p>
$values{'shortabstract'}
</p>
END
    return $result;
}

######################################################################
######################################################################

=pod 

=item &summary_view() 

=cut

######################################################################
######################################################################
sub summary_view {
    my %values = @_;
    my $result=<<END;
<a href="http://$ENV{'HTTP_HOST'}$values{'url'}" 
   target='search_preview'>$values{'author'}</a><br />
$values{'title'}<br />
$values{'owner'} -- $values{'lastrevisiondate'}<br />
$values{'copyrighttag'}<br />
$values{'extrashow'}
</p>
END
    return $result;
}

######################################################################
######################################################################

=pod 

=item &fielded_format_view() 

=cut

######################################################################
######################################################################
sub fielded_format_view {
    my %values = @_;
    my $result=<<END;
<b>URL: </b> <a href="http://$ENV{'HTTP_HOST'}$values{'url'}" 
              target='search_preview'>$values{'url'}</a>
<br />
<b>Title:</b> $values{'title'}<br />
<b>Author(s):</b> $values{'author'}<br />
<b>Subject:</b> $values{'subject'}<br />
<b>Keyword(s):</b> $values{'keywords'}<br />
<b>Notes:</b> $values{'notes'}<br />
<b>MIME Type:</b> $values{'mimetag'}<br />
<b>Language:</b> $values{'language'}<br />
<b>Creation Date:</b> $values{'creationdate'}<br />
<b>Last Revision Date:</b> $values{'lastrevisiondate'}<br />
<b>Publisher/Owner:</b> $values{'owner'}<br />
<b>Copyright/Distribution:</b> $values{'copyrighttag'}<br />
<b>Repository Location:</b> $values{'hostname'}<br />
<b>Abstract:</b> $values{'shortabstract'}<br />
$values{'extrashow'}
</p>
END
    return $result;
}

######################################################################
######################################################################

=pod 

=item &xml_sgml_view() 

=back 

=cut

######################################################################
######################################################################
sub xml_sgml_view {
    my %values = @_;
    my $result=<<END;
<pre>
&lt;LonCapaResource&gt;
&lt;url&gt;$values{'url'}&lt;/url&gt;
&lt;title&gt;$values{'title'}&lt;/title&gt;
&lt;author&gt;$values{'author'}&lt;/author&gt;
&lt;subject&gt;$values{'subject'}&lt;/subject&gt;
&lt;keywords&gt;$values{'keywords'}&lt;/keywords&gt;
&lt;notes&gt;$values{'notes'}&lt;/notes&gt;
&lt;mimeInfo&gt;
&lt;mime&gt;$values{'mime'}&lt;/mime&gt;
&lt;mimetag&gt;$values{'mimetag'}&lt;/mimetag&gt;
&lt;/mimeInfo&gt;
&lt;languageInfo&gt;
&lt;language&gt;$values{'lang'}&lt;/language&gt;
&lt;languagetag&gt;$values{'language'}&lt;/languagetag&gt;
&lt;/languageInfo&gt;
&lt;creationdate&gt;$values{'creationdate'}&lt;/creationdate&gt;
&lt;lastrevisiondate&gt;$values{'lastrevisiondate'}&lt;/lastrevisiondate&gt;
&lt;owner&gt;$values{'owner'}&lt;/owner&gt;
&lt;copyrightInfo&gt;
&lt;copyright&gt;$values{'copyright'}&lt;/copyright&gt;
&lt;copyrighttag&gt;$values{'copyrighttag'}&lt;/copyrighttag&gt;
&lt;/copyrightInfo&gt;
&lt;repositoryLocation&gt;$values{'hostname'}&lt;/repositoryLocation&gt;
&lt;shortabstract&gt;$values{'shortabstract'}&lt;/shortabstract&gt;
&lt;/LonCapaResource&gt;
</pre>
$values{'extrashow'}
END
    return $result;
}

######################################################################
######################################################################

=pod 

=item &filled() see if field is filled.

=cut

######################################################################
######################################################################
sub filled {
    my ($field)=@_;
    if ($field=~/\S/ && $field ne 'any') {
	return 1;
    }
    else {
	return 0;
    }
}

######################################################################
######################################################################

=pod 

=item &output_blank_field_error()

=cut

######################################################################
######################################################################
sub output_blank_field_error {
    my ($r)=@_;
    # make query information persistent to allow for subsequent revision
    my $persistent=&make_persistent(\%ENV);

    $r->print(<<BEGINNING);
<html>
<head>
<title>The LearningOnline Network with CAPA</title>
BEGINNING
    $r->print(<<RESULTS);
</head>
<body bgcolor="#ffffff">
<img align='right' src='/adm/lonIcons/lonlogos.gif' />
<h1>Search Catalog</h1>
<form method="post" action="/adm/searchcat">
$persistent
<input type='button' value='Revise search request'
onClick='this.form.submit();' />
$closebutton
<hr />
<h3>Helpful Message</h3>
<p>
Incorrect search query due to blank entry fields.
You need to fill in the relevant
fields on the search page in order for a query to be
processed.
</p>
</body>
</html>
RESULTS
}

######################################################################
######################################################################

=pod 

=item &output_date_error()

Output a full html page with an error message.

=cut

######################################################################
######################################################################
sub output_date_error {
    my ($r,$message)=@_;
    # make query information persistent to allow for subsequent revision
    my $persistent=&make_persistent(\%ENV);

    $r->print(<<RESULTS);
<html>
<head>
<title>The LearningOnline Network with CAPA</title>
</head>
<body bgcolor="#ffffff">
<img align='right' src='/adm/lonIcons/lonlogos.gif' />
<h1>Search Catalog</h1>
<form method="post" action="/adm/searchcat">
$persistent
<input type='button' value='Revise search request'
onClick='this.form.submit();' />
$closebutton
<hr />
<h3>Helpful Message</h3>
<p>
$message
</p>
</body>
</html>
RESULTS
}

######################################################################
######################################################################

=pod 

=item &start_fresh_session()

Cleans the global %hash by removing all fields which begin with
'pre_' or 'store'.

=cut

######################################################################
######################################################################
sub start_fresh_session {
    delete $hash{'mode_catalog'};
    foreach (keys %hash) {
        if ($_ =~ /^pre_/) {
            delete $hash{$_};
        }
        if ($_ =~ /^store/) {
	    delete $hash{$_};
	}
    }
}

######################################################################
######################################################################

=pod 

=item &popwin_js() send javascript to popwin

=cut

######################################################################
######################################################################
sub popwin_js {
    # Print javascript out to popwin, but make sure we dont generate
    # any javascript errors in doing so.
    my ($r,$text) = @_;
    $r->print(<<"END");
<script type="text/javascript">
    if (! popwin.closed) {
	$text
    }
</script>
END
    $r->rflush();
}

######################################################################
######################################################################

=pod 

=item &popwin_imgupdate()

Send a given image (and its location) out to the browser.  Takes as 
input $r, loncapa server id, and an icon URL.

=cut

######################################################################
######################################################################
sub popwin_imgupdate {
    my ($r,$server,$icon) = @_;
    &popwin_js($r,'popwin.document.img_'.$Apache::lonnet::hostdom{$server}.
               '_'.$server.'.'.'src="/adm/lonIcons/'.$icon.'";');
}    

1;

__END__

=pod

=back 

=cut
